/*
  Copyright (C) 2009 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MAT4_H_
#define MAT4_H_
//----------------------------------------------------------------------------------------------------------------------
/// @file Matrix.h
/// @brief a simple 4x4 Homogenous Affine TX matrix
//----------------------------------------------------------------------------------------------------------------------
// must include types.h first for ngl::Real and GLEW if required
#include "Types.h"

// must include types.h first for ngl::Real and GLEW if required
#include "Types.h"
#include <iosfwd>

namespace ngl
{
class Vec4;
class Quaternion;



//----------------------------------------------------------------------------------------------------------------------
/// @class Mat4 "include/Mat4.h"
/// @brief Matrix Class to do simple matrix operations included
///  operator overloaded functions for maths and matrix * vector type maths.
/// @note for clarity here is the martix offsets m_openGL[n] and the m_00 / [][] versions
/// I tend to draw this out a lot so thought I would put it here
/// to save me doing it again
/// [ (0,0) 0  | (0,1) 1  | (0,2) 2  | (0,3) 3  ]
/// [ (1,0) 4  | (1,1) 5  | (1,2) 6  | (1,3) 7  ]
/// [ (2,0) 8  | (2,1) 9  | (2,2) 10 | (2,3) 11 ]
/// [ (3,0) 12 | (3,1) 13 | (3,2) 14 | (3,3) 15 ]
/// @author Jonathan Macey
/// @version 3.0
/// @date Last Revision 28/09/09 Updated to NCCA Coding standard
/// @example AffineTransforms/Affine.cpp
/// @example AffineTransforms/Global.h
/// @example AffineTransforms/Global.cpp
//----------------------------------------------------------------------------------------------------------------------


class  Mat4
{
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  insertion operator to write matrix to stream
  /// @param[in] output the stream to write to
  /// @param[in] M the matrix to write
  //----------------------------------------------------------------------------------------------------------------------
  friend  std::ostream& operator<<(std::ostream& output, const Mat4& M);
	// added after clang++ build
  friend class Vec4;
public:

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief ctor will always create an identity matrix
  //----------------------------------------------------------------------------------------------------------------------
  Mat4();
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief ctor using 4x4 array, really useful when mixing with Imath as we can do
  /// Imath::Matrix44 <float> iMatrix;
  /// ngl::Mat4 nMatrix(iMatrix.x)
  /// @param[in] _m[4][4] the input array for the matrix
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(Real _m[4][4]);
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief ctor using individual elements
  /// @param [in] _00 0th element (etc you get the deal)
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(
					Real _00,
					Real _01,
					Real _02,
					Real _03,
					Real _10,
					Real _11,
					Real _12,
					Real _13,
					Real _20,
					Real _21,
					Real _22,
					Real _23,
					Real _30,
					Real _31,
					Real _32,
					Real _33

	  );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief copy ctor with reference object
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(
         const Mat4& _m
        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief copy ctor with Real useful for Mat4 m=1; for identity or Matrix m=3.5 for uniform scale
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(
           Real _m
        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief copy ctor with double useful for Mat4 m=1; for identity or Matrix m=3.5 for uniform scale
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(
          double _m
        );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief copy ctor with int useful for Mat4 m=1; for identity or Matrix m=3 for uniform scale
  //----------------------------------------------------------------------------------------------------------------------
  Mat4(
          int _m
        );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set the value at m_m[_x][_y] to _equals
  /// @param[in]  _x the x index into the array
  /// @param[in]  _y the y index into the array
  /// @param[in]  _equals the value to set to
  //----------------------------------------------------------------------------------------------------------------------
  void setAtXY(
                 GLint _x,
                 GLint _y,
                 Real _equals
               );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief clear the matrix to all 0
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& null();

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief  make the matrix m the identity matrix \n
  /// 1 0 0 0 <BR>
  /// 0 1 0 0 <BR>
  /// 0 0 1 0 <BR>
  /// 0 0 0 1 <BR>
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& identity();
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief operator for matrix multiplication
  /// @param[in] _m the matrix to multiply the current one by
  /// @returns this*_m
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 operator*(
                   const Mat4 &_m
                  ) const;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief operator to mult this matrix by value _m
  /// @param[in] _m the matrix to multiplt
  /// @returns a new matrix this*_m
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& operator*=(
                           const Mat4 &_m
                          );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief operator to add two matrices together
  /// @param[in] _m the matrix to add
  /// @returns this+_m
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 operator+(
                    const Mat4 &_m
                  ) const;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief += operator
  /// @param[in] _m the matrix to add
  /// @returns this+m
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& operator+=(
                            const Mat4 &_m
                          );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief operator to mult matrix by a scalar
  /// @param[in] _i the scalar to multiply by
  /// @returns this*_i
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 operator*(
                    const Real _i
                  ) const;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief *= operator with a scalar value
  /// @param[in] _i the scalar to multiply by
  /// @returns the matrix*i
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& operator*=(
                            const Real _i
                          );
  //----------------------------------------------------------------------------------------------------------------------
  ///  @brief method to transpose the matrix
  //----------------------------------------------------------------------------------------------------------------------
  const Mat4& transpose();

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set this matrix to a rotation matrix in the X axis for value _deg
  /// note the matrix should be set to identity before doing this
  /// @param[in] _deg the value to be rotated by in degrees
  //----------------------------------------------------------------------------------------------------------------------
  void rotateX(
                const Real _deg
              );
  /// @brief set this matrix to a rotation matrix in the Y axis for value _deg
  /// note the matrix should be set to identity before doing this
  /// @param[in] _deg the value to be rotated by in degrees

  void rotateY(
               const Real _deg
              );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set this matrix to a rotation matrix in the Z axis for value _deg
  /// note the matrix should be set to identity before doing this
  /// @param[in] _deg the value to be rotated by in degrees
  //----------------------------------------------------------------------------------------------------------------------
   void rotateZ(
                 const Real _deg
                 );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set the matrix scale values
  /// @param[in] _x the scale value in the _x
  /// @param[in] _y the scale value in the _y
  /// @param[in] _z the scale value in the _z
  //----------------------------------------------------------------------------------------------------------------------
  void scale(
             const Real _x,
             const Real _y,
             const Real _z
            );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief set the matrix as a translation matrix
  /// @param[in] _x the _x translation value
  /// @param[in] _y the _y translation value
  /// @param[in] _z the _z translation value
  //----------------------------------------------------------------------------------------------------------------------
  void translate(
                 const Real _x,
                 const Real _y,
                 const Real _z
                );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief get the determinant of the matrix
  /// @returns the determinat
  //----------------------------------------------------------------------------------------------------------------------
  Real determinant() const;

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief get a sub 3x3 matrix used in determinant and Inverse calculations
  /// @param[in] _i index value
  /// @param[in] _j index value
  /// @param[in] o_mat the 3x3 matrix
  //----------------------------------------------------------------------------------------------------------------------
  void subMatrix3x3(
                    const int _i,
                    const int _j,
                    Real o_mat[]
                   ) const;


  Real * as3x3Array() const;

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief get the inverse of the matrix
  /// @returns a new matrix the inverse of the current matrix (warning no error checking )
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 inverse();
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief convert this matrix to a Quaternion
  /// @returns the matrix as a Quaternion
  //----------------------------------------------------------------------------------------------------------------------
  Quaternion asQuaternion() const;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief multiply this by a vector
  /// @param[in] _v the vector to multiply
  /// @returns Vector M*V
  //----------------------------------------------------------------------------------------------------------------------
  Vec4 operator * (
                     const Vec4 &_v
                    ) const;
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief axis / angle rotation using the Euler method
  /// @param[in] _angle the angle in degree to rotate
  /// @param[in] _x part of the axis should be normalised before call
  /// @param[in] _y part of the axis should be normalised before call
  /// @param[in] _z part of the axis  should be normalised before call
  //----------------------------------------------------------------------------------------------------------------------
  void euler(
             const Real _angle,
             const Real _x,
             const Real _y,
             const Real _z
            );
  //----------------------------------------------------------------------------------------------------------------------
  /// @brief accesor to the openGL matrix
  /// @returns a pointer to m_openGL[0]
  //----------------------------------------------------------------------------------------------------------------------
  inline Real * openGL(){return &m_openGL[0];}

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief returns a matrix adjacent to the matrix passed in
  /// @returns the adjacent matrix
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 Adjacent( const ngl::Mat4 &_mat );

  //----------------------------------------------------------------------------------------------------------------------
  /// @brief returns the  ajacent matrix to this
  /// @returns the adjacent matrix
  //----------------------------------------------------------------------------------------------------------------------
  Mat4 Adjacent( );

public :
 //----------------------------------------------------------------------------------------------------------------------
 // allow access to some of the other classes
 friend class Transformation;
 friend class Quaternion;
 friend class Camera;
 //----------------------------------------------------------------------------------------------------------------------

#ifndef BUILDING_DOCS

  union
  {
#endif
    //----------------------------------------------------------------------------------------------------------------------
    /// @brief matrix element m_m as a 4x4 array mapped by union to m_nn elements and m_openGL
    //----------------------------------------------------------------------------------------------------------------------
    Real m_m[4][4];
    //----------------------------------------------------------------------------------------------------------------------
    /// @brief  The matrix in m_openGL 16 Real array format usefull for OpenGL fv formats
    /// mapped to m_m[][] elements and m_xx elements
    //----------------------------------------------------------------------------------------------------------------------
    Real m_openGL[16];

#ifndef BUILDING_DOCS

    struct
    {
#endif
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[0][0] or m_openGL[0]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_00;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[0][1] or m_openGL[1]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_01;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[0][2] or m_openGL[2]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_02;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[0][3] or m_openGL[3]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_03;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[1][0] or m_openGL[4]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_10;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[1][1] or m_openGL[5]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_11;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[1][2] or m_openGL[6]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_12;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[1][3] or m_openGL[7]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_13;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[2][0] or m_openGL[8]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_20;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[2][1] or m_openGL[9]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_21;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[2][2] or m_openGL[10]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_22;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[2][3] or m_openGL[11]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_23;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[3][0] or m_openGL[12]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_30;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[3][1] or m_openGL[13]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_31;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[3][2] or m_openGL[14]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_32;
      //----------------------------------------------------------------------------------------------------------------------
      /// @brief  individual matrix element maps to m_m[3][3] or m_openGL[15]
      //----------------------------------------------------------------------------------------------------------------------
      Real m_33;
      #ifndef BUILDING_DOCS
    };

   };
#endif

  }; // end of class
}// end of namespace

#endif
//----------------------------------------------------------------------------------------------------------------------

